__author__ = "shihhau"  # Shih Hau

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pandas

from findatapy.market.datavendor import DataVendor
from findatapy.util import LoggerManager


class DataVendorBitcoincharts(DataVendor):
    """Class for reading in data from various web sources into findatapy
    library including
    """

    def __init__(self):
        super(DataVendorBitcoincharts, self).__init__()

    # implement method in abstract superclass
    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)
        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request data from Bitcoincharts")

        data_website = 'http://api.bitcoincharts.com/v1/csv/' + \
                       md_request_vendor.tickers[0] + '.csv.gz'
        data_frame = pandas.read_csv(data_website,
                                     names=['datetime', 'close', 'volume'])
        data_frame = data_frame.set_index('datetime')
        data_frame.index = pandas.to_datetime(data_frame.index, unit='s')
        data_frame.index.name = 'Date'
        data_frame = data_frame[
            (data_frame.index >= md_request_vendor.start_date) & (
                    data_frame.index <= md_request_vendor.finish_date)]
        #        data_frame = df[~df.index.duplicated(keep='last')]
        if len(data_frame) == 0:
            logger.warning(
                "Warning: No data. Please change the start_date and finish_date.")

        data_frame.columns = [md_request.tickers[0] + '.close',
                              md_request.tickers[0] + '.volume']
        logger.info("Completed request from Bitcoincharts.")

        return data_frame


###############################################################################

class DataVendorPoloniex(DataVendor):
    """Class for reading in data from various web sources into findatapy
    library including
    """

    def __init__(self):
        super(DataVendorPoloniex, self).__init__()

    # Implement method in abstract superclass
    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)

        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request data from Poloniex")

        poloniex_url = 'https://poloniex.com/public?command=returnChartData&currencyPair={}&start={}&end={}&period={}'

        if md_request_vendor.freq == 'intraday':
            period = 300
        if md_request_vendor.freq == 'daily':
            period = 86400

        json_url = poloniex_url.format(
            md_request_vendor.tickers[0],
            int(md_request_vendor.start_date.timestamp()),
            int(md_request_vendor.finish_date.timestamp()),
            period)
        data_frame = pandas.read_json(json_url)
        data_frame = data_frame.set_index('date')
        data_frame.index.name = 'Date'

        if data_frame.index[0] == 0:
            logger.warning(
                "Warning: No data. Please change the start_date and finish_date.")

        data_frame.columns = [md_request.tickers[0] + '.close',
                              md_request.tickers[0] + '.high',
                              md_request.tickers[0] + '.low',
                              md_request.tickers[0] + '.open',
                              md_request.tickers[0] + '.quote-volume',
                              md_request.tickers[0] + '.volume',
                              md_request.tickers[
                                  0] + '.weighted-average']

        field_selected = []
        for i in range(0, len(md_request_vendor.fields)):
            field_selected.append(0)
            field_selected[-1] = md_request.tickers[0] + '.' + \
                                 md_request_vendor.fields[i]

        logger.info("Completed request from Poloniex")

        return data_frame[field_selected]


###############################################################################

class DataVendorBinance(DataVendor):
    """Class for reading in data from various web sources into findatapy
    library including
    """

    # Data limit = 500

    def __init__(self):
        super(DataVendorBinance, self).__init__()

    # implement method in abstract superclass
    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)

        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request data from Binance")

        import time
        binance_url = 'https://www.binance.com/api/v1/klines?symbol={}&interval={}&startTime={}&endTime={}'
        if md_request_vendor.freq == 'intraday':
            period = '1m'
        if md_request_vendor.freq == 'daily':
            period = '1d'

        data_frame = pandas.DataFrame(columns=[0, 1, 2, 3, 4, 5])
        start_time = int(
            md_request_vendor.start_date.timestamp() * 1000)
        finish_time = int(
            md_request_vendor.finish_date.timestamp() * 1000)

        stop_flag = 0
        while stop_flag == 0:
            if stop_flag == 1:
                break
            json_url = binance_url.format(
                md_request_vendor.tickers[0], period, start_time,
                finish_time)
            data_read = pandas.read_json(json_url)

            if len(data_read) < 500:
                if len(data_read) == 0 & len(data_frame) == 0:
                    logger.warning(
                        "Warning: No data. Please change the start_date and finish_date.")
                    break
                else:
                    stop_flag = 1
            data_frame = data_frame.append(data_read)
            start_time = int(data_frame[0].tail(1))
            time_library.sleep(2)

        if (len(data_frame) == 0):
            return data_frame

        data_frame.columns = ['open-time', 'open', 'high', 'low', 'close',
                              'volume', 'close-time', 'quote-asset-volume',
                              'trade-numbers', 'taker-buy-base-asset-volume',
                              'taker-buy-quote-asset-volume', 'ignore']
        data_frame['open-time'] = data_frame['open-time'] / 1000
        data_frame = data_frame.set_index('open-time')
        data_frame = data_frame.drop(['close-time', 'ignore'], axis=1)
        data_frame.index.name = 'Date'
        data_frame.index = pandas.to_datetime(data_frame.index, unit='s')
        data_frame.columns = [md_request.tickers[0] + '.open',
                              md_request.tickers[0] + '.high',
                              md_request.tickers[0] + '.low',
                              md_request.tickers[0] + '.close',
                              md_request.tickers[0] + '.volume',
                              md_request.tickers[
                                  0] + '.quote-asset-volume',
                              md_request.tickers[
                                  0] + '.trade-numbers',
                              md_request.tickers[
                                  0] + '.taker-buy-base-asset-volume',
                              md_request.tickers[
                                  0] + '.taker-buy-quote-asset-volume']

        field_selected = []
        for i in range(0, len(md_request_vendor.fields)):
            field_selected.append(0)
            field_selected[-1] = md_request.tickers[0] + '.' + \
                                 md_request_vendor.fields[i]

        logger.info("Completed request from Binance")

        return data_frame[field_selected]


###############################################################################

class DataVendorBitfinex(DataVendor):
    """Class for reading in data from various web sources into findatapy
    library including
    """

    # Data limit = 1000

    def __init__(self):
        super(DataVendorBitfinex, self).__init__()

    # implement method in abstract superclass
    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)
        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request data from Bitfinex.")

        import time
        bitfinex_url = 'https://api.bitfinex.com/v2/candles/trade:{}:t{}/hist?start={}&end={}&limit=1000&sort=1'
        if md_request_vendor.freq == 'intraday':
            period = '1m'
        if md_request_vendor.freq == 'daily':
            period = '1D'

        data_frame = pandas.DataFrame(columns=[0, 1, 2, 3, 4, 5])
        start_time = int(
            md_request_vendor.start_date.timestamp() * 1000)
        finish_time = int(
            md_request_vendor.finish_date.timestamp() * 1000)
        stop_flag = 0
        while stop_flag == 0:
            if stop_flag == 1:
                break
            json_url = bitfinex_url.format(period,
                                           md_request_vendor.tickers[
                                               0], start_time, finish_time)
            data_read = pandas.read_json(json_url)
            if (len(data_read) < 1000):
                if ((len(data_read) == 0) & (len(data_frame) == 0)):
                    break
                else:
                    stop_flag = 1
            data_frame = data_frame.append(data_read)
            start_time = int(data_frame[0].tail(1))
            time_library.sleep(2)

        if len(data_frame) == 0:
            logger.warning(
                "Warning: No data. Please change the start_date and finish_date.")

        #    return data_frame

        data_frame.columns = ['mts', 'open', 'close', 'high', 'low', 'volume']
        data_frame = data_frame.set_index('mts')

        data_frame = data_frame[~data_frame.index.duplicated(keep='first')]

        data_frame.index.name = 'Date'
        data_frame.index = pandas.to_datetime(data_frame.index, unit='ms')
        data_frame.columns = [md_request.tickers[0] + '.open',
                              md_request.tickers[0] + '.close',
                              md_request.tickers[0] + '.high',
                              md_request.tickers[0] + '.low',
                              md_request.tickers[0] + '.volume']

        field_selected = []
        for i in range(0, len(md_request_vendor.fields)):
            field_selected.append(0)
            field_selected[-1] = md_request.tickers[0] + '.' + \
                                 md_request_vendor.fields[i]

        logger.info("Completed request from Bitfinex.")

        return data_frame[field_selected]


###############################################################################

class DataVendorGdax(DataVendor):
    """Class for reading in data from various web sources into findatapy
    library including
    """

    # Data limit = 350

    def __init__(self):
        super(DataVendorGdax, self).__init__()

    # implement method in abstract superclass
    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)

        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request data from Gdax.")

        gdax_url = 'https://api.gdax.com/products/{}/candles?start={}&end={}&granularity={}'
        start_time = md_request_vendor.start_date
        end_time = md_request_vendor.finish_date
        if md_request_vendor.freq == 'intraday':
            # 1 minute data
            period = '60'
            dt = timedelta(minutes=1)
        if md_request_vendor.freq == 'daily':
            period = '86400'
            dt = timedelta(days=1)
        limit = 350

        data_frame = pandas.DataFrame(columns=[0, 1, 2, 3, 4, 5])
        stop_flag = 0
        while stop_flag == 0:
            if stop_flag == 1:
                break
            data_end_time = start_time + (limit - 1) * dt
            if data_end_time > end_time:
                data_end_time = end_time
                stop_flag = 1
            json_url = gdax_url.format(md_request_vendor.tickers[0],
                                       start_time.isoformat(),
                                       data_end_time.isoformat(), period)
            data_read = pandas.read_json(json_url)
            data_frame = data_frame.append(data_read)
            if (len(data_read) == 0):
                start_time = data_end_time
            else:
                start_time = pandas.to_datetime(int(data_read[0].head(1)),
                                                unit='s')
            time_library.sleep(2)

        if len(data_frame) == 0:
            logger.warning(
                "Warning: No data. Please change the start_date and finish_date.")

        data_frame.columns = ['time', 'low', 'high', 'open', 'close', 'volume']
        data_frame = data_frame.set_index('time')
        data_frame.index = pandas.to_datetime(data_frame.index, unit='s')
        data_frame.index.name = 'Date'
        data_frame = data_frame[~data_frame.index.duplicated(keep='first')]
        data_frame = data_frame.sort_index(ascending=True)
        data_frame.columns = [md_request.tickers[0] + '.low',
                              md_request.tickers[0] + '.high',
                              md_request.tickers[0] + '.open',
                              md_request.tickers[0] + '.close',
                              md_request.tickers[0] + '.volume']

        field_selected = []
        for i in range(0, len(md_request_vendor.fields)):
            field_selected.append(0)
            field_selected[-1] = md_request.tickers[0] + '.' + \
                                 md_request_vendor.fields[i]

        logger.info("Completed request from Gdax.")

        return data_frame[field_selected]


###############################################################################

class DataVendorKraken(DataVendor):
    """Class for reading in data from various web sources into findatapy
    library including
    """

    # Data limit : can only get the most recent 720 rows for klines
    # Collect data from all trades data

    def __init__(self):
        super(DataVendorKraken, self).__init__()

    # implement method in abstract superclass
    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)

        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request data from Kraken.")

        # kraken_url = 'https://api.kraken.com/0/public/OHLC?pair={}&interval={}&since=0'
        # if md_request_vendor.freq == 'intraday':
        #    period = 1
        # if md_request_vendor.freq == 'daily':
        #    period = 1440
        start_time = int(
            md_request_vendor.start_date.timestamp() * 1e9)
        end_time = int(
            md_request_vendor.finish_date.timestamp() * 1e9)

        kraken_url = 'https://api.kraken.com/0/public/Trades?pair={}&since={}'
        data_frame = pandas.DataFrame(
            columns=['close', 'volume', 'time', 'buy-sell', 'market-limit',
                     'miscellaneous'])
        stop_flag = 0

        while stop_flag == 0:
            if stop_flag == 1:
                break

            json_url = kraken_url.format(md_request_vendor.tickers[0],
                                         start_time)
            data_read = json.loads(requests.get(json_url).text)
            if (len(list(data_read)) == 1):
                time_library.sleep(10)
                data_read = json.loads(requests.get(json_url).text)

            data_list = list(data_read['result'])[0]
            data_read = data_read['result'][data_list]
            df = pandas.DataFrame(data_read,
                                  columns=['close', 'volume', 'time',
                                           'buy-sell', 'market-limit',
                                           'miscellaneous'])
            start_time = int(df['time'].tail(1) * 1e9)
            if (start_time > end_time):
                stop_flag = 1
            if (end_time < int(df['time'].head(1) * 1e9)):
                stop_flag = 1
            data_frame = data_frame.append(df)
            time_library.sleep(5)

        data_frame = data_frame.set_index('time')
        data_frame.index = pandas.to_datetime(data_frame.index, unit='s')
        data_frame.index.name = 'Date'
        data_frame = data_frame.drop(['miscellaneous'], axis=1)
        data_frame.replace(['b', 's', 'm', 'l'], [1, -1, 1, -1], inplace=True)
        data_frame = data_frame[
            (data_frame.index >= md_request_vendor.start_date) & (
                    data_frame.index <= md_request_vendor.finish_date)]
        if len(data_frame) == 0:
            logger.warning(
                "Warning: No data. Please change the start_date and finish_date.")

        data_frame.columns = [md_request.tickers[0] + '.close',
                              md_request.tickers[0] + '.volume',
                              md_request.tickers[0] + '.buy-sell',
                              md_request.tickers[0] + '.market-limit']

        field_selected = []
        for i in range(0, len(md_request_vendor.fields)):
            field_selected.append(0)
            field_selected[-1] = md_request.tickers[0] + '.' + \
                                 md_request_vendor.fields[i]

        logger.info("Completed request from Kraken.")

        return data_frame[field_selected]


###############################################################################

class DataVendorBitmex(DataVendor):
    """Class for reading in data from various web sources into findatapy
    library including
    """

    # Data limit = 500,  150 calls / 5 minutes

    def __init__(self):
        super(DataVendorBitmex, self).__init__()

    # implement method in abstract superclass
    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)
        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request data from Bitmex.")

        bitMEX_url = 'https://www.bitmex.com/api/v1/quote?symbol={}&count=500&reverse=false&startTime={}&endTime={}'
        data_frame = pandas.DataFrame(
            columns=['askPrice', 'askSize', 'bidPrice', 'bidSize', 'symbol',
                     'timestamp'])
        start_time = md_request_vendor.start_date.timestamp()
        finish_time = md_request_vendor.finish_date.timestamp()
        symbol = md_request_vendor.tickers[0]

        stop_flag = 0
        while stop_flag == 0:
            if stop_flag == 1:
                break
            json_url = bitMEX_url.format(symbol, start_time.isoformat(),
                                         finish_time.isoformat())
            data_read = pandas.read_json(json_url)
            if (len(data_read) < 500):
                stop_flag = 1
            data_frame = data_frame.append(data_read)
            start_time = data_read['timestamp'][data_frame.index[-1]]
            time_library.sleep(2)

        if (len(data_frame) == 0):
            logger.warning(
                "Warning: No data. Please change the start_date and finish_date.")

        data_frame = data_frame.drop(columns=['symbol'])
        col = ['ask-price', 'ask-size', 'bid-price', 'bid-size', 'timestamp']
        data_frame.columns = col
        data_frame = data_frame.set_index('timestamp')
        data_frame.index = pandas.to_datetime(data_frame.index, unit='ms')
        data_frame = data_frame[~data_frame.index.duplicated(keep='first')]
        data_frame.columns = [md_request.tickers[0] + '.ask-price',
                              md_request.tickers[0] + '.ask-size',
                              md_request.tickers[0] + '.bid-price',
                              md_request.tickers[0] + '.bid-size']

        field_selected = []
        for i in range(0, len(md_request_vendor.fields)):
            field_selected.append(0)
            field_selected[-1] = md_request.tickers[0] + '.' + \
                                 md_request_vendor.fields[i]

        logger.info("Completed request from Bitfinex.")

        return data_frame[field_selected]


class DataVendorHuobi(DataVendor):
    """Class for reading in data from various web sources into findatapy
    library including
    """

    # Data limit = 500,  150 calls / 5 minutes

    def __init__(self):
        super(DataVendorHuobi, self).__init__()

    # implement method in abstract superclass
    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)

        def _calc_period_size(freq, start_dt, finish_dt):
            actual_window = finish_dt - start_dt
            extra_window = datetime.datetime.now() - finish_dt
            request_window = actual_window + extra_window

            if freq == 'daily':
                return int(request_window.days), '1day'

            if freq == 'tick':
                request_minutes = request_window.total_seconds() / 60
                return int(request_minutes), '1min'

            raise ValueError("Unsupported freq: '{}'".format(freq))

        # need to trick huobi to think we are a web-browser
        header = {
            "User-Agent": "Mozilla/5.0",
            "X-Requested-With": "XMLHttpRequest"
        }

        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request data from Huobi.")

        request_size, period = _calc_period_size(
            md_request_vendor.freq,
            md_request_vendor.start_date,
            md_request_vendor.finish_date)

        if request_size > 2000:
            raise ValueError(
                "Requested data too old for candle-stick frequency of '{}'".
                format(
                    md_request_vendor.freq))

        url = "https://api.huobi.pro/market/history/kline?period={period}&size={size}&symbol={symbol}".format(
            period=period,
            size=request_size,
            symbol=md_request_vendor.tickers[0]
        )

        response = requests.get(url, headers=header)
        raw_data = json.loads(response.text)
        df = pandas.DataFrame(raw_data["data"])
        df["timestamp"] = pandas.to_datetime(df["id"], unit="s")

        df = df.set_index("timestamp").sort_index(ascending=True)
        df = df[~df.index.duplicated(keep='first')]

        df.drop(["id"], axis=1, inplace=True)

        if df.empty:
            logger.info(
                "Warning: No data. Please change the start_date "
                "and finish_date.")

        df.columns = ["{}.{}".format(md_request.tickers[0], col) for
                      col in df.columns]

        field_selected = []
        for i in range(0, len(md_request_vendor.fields)):
            field_selected.append(0)
            field_selected[-1] = md_request.tickers[0] + '.' + \
                                 md_request_vendor.fields[i]

        logger.info("Completed request from Huobi.")

        df = df[field_selected]
        return df


__author__ = "saeedamen"  # Saeed Amen

#
# Copyright 2024 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#

import datetime
import time

if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool;

    SwimPool()

    # This is the new ArcticDB, which uses disk based storage, as opposed
    # to a database engine like MongoDB as a backend
    from findatapy.market import Market, MarketDataRequest, \
        MarketDataGenerator, IOEngine

    market = Market(market_data_generator=MarketDataGenerator())

    # In the config file, we can use keywords 'open', 'high', 'low', 'close'
    # and 'volume' for Yahoo data

    # Download equities data from Yahoo
    md_request = MarketDataRequest(
        start_date="01 Mar 2020",  # start date
        data_source="yahoo",  # use yahoo as data source
        tickers=["Apple", 'Nvidia'],  # ticker (findatapy)
        fields=["close"],  # which fields to download
        vendor_tickers=["aapl", "nvda"],  # ticker (Yahoo)
        vendor_fields=["Close"])  # which Yahoo fields to download)

    df = market.fetch_market(md_request)

    io = IOEngine()

    arcticdb_conn_str = "arcticdb:lmdb://tempdatabase?map_size=2GB"

    # To switch between local storage or s3, it's a matter of changing the
    # connection string (also you need to make sure your AWS S3 authentication is set etc.)
    if local_storage:
        arcticdb_conn_str = "arcticdb:lmdb://tempdatabase?map_size=2GB"
    else:
        # https://docs.arcticdb.io/latest/#s3-configuration gives more details
        # Note we need to prefix arcticdb: to the front so findatapy
        # knows what backend engine to use
        region = "eu-west-2"
        bucket_name = "burger_king_whopper"  # Not sure, if this name is taken :-)
        path_prefix = "test"

        arcticdb_conn_str = f"arcticdb:s3s://s3.{region}.amazonaws.com:{bucket_name}?path_prefix={path_prefix}&aws_auth=true"

    # Set various parameters to govern how we write to ArcticDB, to use
    # versioning
    arcticdb_dict = {
        # If this is set to true removes previous versions (so we only record
        # the final version). Not pruning versions will take more disk space
        "prune_previous_versions": False,

        # Do we want to append to existing records or write
        # If you attempt to append with an overlapping chunk, you'll
        # get an assertion failure, "update" allows you to change existing data
        "write_style": "write",  # "write" / "append" / "update"

        # If set to true will remove any existing library, before writing
        "force_create_library": False,

        # This enables us to take advantage of ArcticDB's filtering of columns/dates
        # otherwise we would download the full dataset, and then filter
        # in Pandas
        "allow_on_disk_filter": True,

        # You can also specify your own custom queries for ArcticDB
        "query_builder": None
    }

    current_time_list = []

    for i in range(0, 3):

        # Perturb by a small amount to differientate time series versions
        df_to_write = df * (1 + i / 100)

        # Note: we will use the lmdb interface which is for local filestorage
        # with arcticdb (it also supports s3, azure etc.)
        io.write_time_series_cache_to_disk(
            "stocks", df_to_write, engine=arcticdb_conn_str,
            arcticdb_dict=arcticdb_dict)

        current_time = datetime.datetime.now().utcnow()
        current_time_list.append(current_time)
        time.sleep(5)

    for i in range(0, 3):
        # Read back from Arctic
        df_arcticdb = io.read_time_series_cache_from_disk(
            "stocks", engine=arcticdb_conn_str,
            arcticdb_dict=arcticdb_dict,
            as_of=current_time_list[i])

        print(f"Read version as of {str(current_time_list[i])}")
        print(df_arcticdb.tail(n=5))

    print("Read without as of parameter - will return the last version")

    # If we do not specific the as_of, we get the final version
    df_arcticdb = io.read_time_series_cache_from_disk(
        "stocks", engine=arcticdb_conn_str,
        arcticdb_dict=arcticdb_dict)

    print(df_arcticdb.tail(n=5))



import pytest
import pandas as pd
import itertools

from findatapy.market import Market, MarketDataGenerator, MarketDataRequest
from findatapy.util.dataconstants import DataConstants
from typing import Optional

market = Market(market_data_generator=MarketDataGenerator())

data_constants = DataConstants()
quandl_api_key = data_constants.quandl_api_key


def test_quandl_download():
    md_request = MarketDataRequest(start_date="year", category="fx",
                                   data_source="quandl", tickers=["AUDJPY"],
                                   quandl_api_key=quandl_api_key)

    df = market.fetch_market(md_request)

    assert df is not None


def test_yahoo_download():
    md_request = MarketDataRequest(
        start_date="year",  # start date
        data_source="yahoo",  # use Bloomberg as data source
        tickers=["Apple", "Citigroup"],  # ticker (findatapy)
        fields=["close"],  # which fields to download
        vendor_tickers=["aapl", "c"],  # ticker (Yahoo)
        vendor_fields=["Close"])  # which Bloomberg fields to download)

    df = market.fetch_market(md_request)

    assert df is not None

def test_bbg_download():

    freq = ["daily", "intraday"]

    for fr in freq:
        md_request = MarketDataRequest(
            start_date="week",  # start date
            data_source="bloomberg",  # use Bloomberg as data source
            freq=fr,
            tickers=["S&P 500", "EURUSD"],  # ticker (findatapy)
            fields=["close"],  # which fields to download
            vendor_tickers=["SPX Index", "EURUSD Curncy"],  # ticker (Yahoo)
            vendor_fields=["PX_LAST"])  # which Bloomberg fields to download)

        df = market.fetch_market(md_request)

        assert df is not None


def test_free_fx_tick_download():
    # First we can do it by defining all the vendor fields, tickers etc. 
    # so we bypass the configuration file
    md_request = MarketDataRequest(start_date="05 Dec 2016", 
                                   finish_date="07 Dec 2016",
                                   fields=["bid"], vendor_fields=["bid"],
                                   freq="tick", data_source="dukascopy",
                                   tickers=["EURUSD"], 
                                   vendor_tickers=["EURUSD"])

    md_request.data_source = "dukascopy"
    df = market.fetch_market(md_request).resample("1min").mean()

    assert df is not None

class MarketFetchTestRunner:

    def test_df_columns_return(self, md_request: MarketDataRequest, data_source_overwrite: Optional[str]=None):
        df = self.get_market_df(md_request, data_source_overwrite)
        for field in self.get_price_fields(md_request):
            assert field in df.columns

    def test_start_time_post_given_start_time(self, md_request: MarketDataRequest, data_source_overwrite: Optional[str]=None):
        df = self.get_market_df(md_request, data_source_overwrite)
        first_tick = df.index.tolist()[0]
        first_tick = first_tick.replace(tzinfo=None)
        assert first_tick >= md_request.start_date

    def test_finish_time_pre_given_finish_time(self, md_request: MarketDataRequest, data_source_overwrite: Optional[str]=None):
        df = self.get_market_df(md_request, data_source_overwrite)
        last_tick = df.index.tolist()[-1]
        last_tick = last_tick.replace(tzinfo=None)
        assert last_tick <= md_request.finish_date
 
    def test_valid_prices(self, md_request: MarketDataRequest, data_source_overwrite: Optional[str]=None):
        df = self.get_market_df(md_request, data_source_overwrite)
        quarter_null_fields = []
        for field in self.get_price_fields(md_request):
            if df[field].isnull().mean() > 0.25:
                quarter_null_fields.append(field)
        assert [] == quarter_null_fields
        
    def get_market_df(self, md_request: MarketDataRequest, data_source_overwrite: Optional[str]=None):
        md_request.data_source = data_source_overwrite if data_source_overwrite is not None else md_request.data_source
        df = market.fetch_market(md_request)
        assert isinstance(df, pd.DataFrame)
        return df
    
    def get_price_fields(self, md_request: MarketDataRequest):
        all_fields = ["high", "low", "close", "volume"]
        desired_fields = md_request.fields if md_request.fields is not None else all_fields
        fields = [f"{i[0]}.{i[1]}" for i in list(itertools.product(md_request.tickers, desired_fields))]
        return fields

md_requests = [
    # I dont have bloomberg or other keys to test...            
    MarketDataRequest(  
        start_date="year",  # start date
        data_source="yahoo",  # use Bloomberg as data source
        tickers=["Apple", "Citigroup"],  # ticker (findatapy)
        # fields=["close"],  # which fields to download
        vendor_tickers=["aapl", "c"],  # ticker (Yahoo)
        vendor_fields=["Close"],  # which Bloomberg fields to download)
        cache_algo='internet_load_return',  # Disable caching
        push_to_cache=False
    ),
    MarketDataRequest(start_date="05 Dec 2016", 
        finish_date="07 Dec 2016",
        fields=["bid"], vendor_fields=["bid"],
        freq="tick", data_source="dukascopy",
        tickers=["EURUSD"], 
        vendor_tickers=["EURUSD"]
    ),
]

fetch_runner = MarketFetchTestRunner()

@pytest.mark.parametrize("md_request", md_requests)
def test_df_columns(md_request: MarketDataRequest):
    fetch_runner.test_df_columns_return(md_request)

@pytest.mark.parametrize("md_request", md_requests)
def test_start_dates(md_request: MarketDataRequest):
    fetch_runner.test_start_time_post_given_start_time(md_request)

@pytest.mark.parametrize("md_request", md_requests)
def test_finish_dates(md_request: MarketDataRequest):
    fetch_runner.test_finish_time_pre_given_finish_time(md_request)

@pytest.mark.parametrize("md_request", md_requests)
def test_price_fields_not_quarter_null(md_request: MarketDataRequest):
    fetch_runner.test_valid_prices(md_request)

if __name__ == "__main__":
    pytest.main()




__author__ = "saeedamen"  # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import pandas as pd

from findatapy.market.ioengine import IOEngine

from findatapy.util.dataconstants import DataConstants

data_constants = DataConstants()

redis_server = data_constants.db_cache_server
redis_port = data_constants.db_cache_port

def test_redis_caching():
    # Note: you need to install Redis in order for this to work!

    # read CSV from disk, and make sure to parse dates
    df = pd.read_csv("S&P500.csv", parse_dates=['Date'], index_col=['Date'])
    df.index = pd.to_datetime(df.index)

    io = IOEngine()

    use_cache_compression = [True, False]

    for u in use_cache_compression:
        # Write DataFrame to Redis (using pyarrow format)
        io.write_time_series_cache_to_disk('test_key', df, engine='redis', db_server=redis_server, db_port=redis_port,
                                           use_cache_compression=u)

        # Read back DataFrame from Redis (using pyarrow format)
        df_out = io.read_time_series_cache_from_disk('test_key', engine='redis', db_server=redis_server, db_port=redis_port)

        pd.testing.assert_frame_equal(df, df_out)


def test_path_join():

    io = IOEngine()

    path = io.path_join("/home/hello", "hello", "hello")

    assert path == "/home/hello/hello/hello"

    path = io.path_join("s3://home/hello", "hello", "hello")

    assert path == "s3://home/hello/hello/hello"

if __name__ == '__main__':
    pytest.main()



__author__ = "saeedamen"  # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#


if __name__ == "__main__":
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing 
    # doesn"t work! (not necessary on Linux)
    from findatapy.util import SwimPool;

    SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    # choose run_example = 0 for everything
    # run_example = 1 - download minute crypto data from Bloomberg
    # coming soon other crypto data sources!

    run_example = 1

    if run_example == 1 or run_example == 0:
        ####### Bloomberg

        # Now we will try various examples with Bloomberg
        # only works if you have Bloomberg terminal installed and the 
        # Python API!
        md_request = MarketDataRequest(start_date="week", freq="intraday",
                                       category="fx", data_source="bloomberg",
                                       cut="BSTP",
                                       tickers=["XBTUSD", "XETUSD", "XLCUSD",
                                                "XRPUSD"])

        df = market.fetch_market(md_request)

        print(df.tail(n=10))




from findatapy.util import DataConstants

databento_api_key = DataConstants().databento_api_key
import databento as db

client = db.Historical(databento_api_key)

# RYZ4 and RPZ4 (EURJPY and EURGBP futures respectively)

data = client.timeseries.get_range(
    dataset="GLBX.MDP3",
    symbols=["RYZ4", "RPZ4"],
    schema="ohlcv-1d",
    start="2024-10-01T00:00:00",
    end="2024-11-08T00:10:00",
    # limit=1,
)
df = data.to_df()
df = df.set_index('symbol', append=True).unstack(level='symbol')
fields = df.columns.levels[0]
tickers = df.columns.levels[1]

new_cols = []

for fi in fields:
    for ti in tickers:
        new_cols.append(ti + "." + fi)

df.columns = new_cols
df.index.name = "Date"

print(df)


print(df.iloc[0].to_json(indent=4))


__author__ = "saeedamen"  # Saeed Amen

#
# Copyright 2022 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import pandas as pd

from findatapy.timeseries import Filter


def test_filtering_by_dates():

    filter = Filter()

    # filter S&P500 between specific working days
    start_date = '01 Oct 2008'
    finish_date = '29 Oct 2008'

    # read CSV from disk, and make sure to parse dates
    df = pd.read_csv("S&P500.csv", parse_dates=['Date'], index_col=['Date'])
    df = filter.filter_time_series_by_date(start_date=start_date,
                                           finish_date=finish_date,
                                           data_frame=df)

    assert df.index[0] == pd.to_datetime(start_date)
    assert df.index[-1]== pd.to_datetime(finish_date)


if __name__ == '__main__':
    pytest.main()



__author__ = "saeedamen"  # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing
    # doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool;

    SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    market = Market(market_data_generator=MarketDataGenerator())

    # choose run_example = 0 for everything
    # run_example = 1 - download implied volatility data from Bloomberg for FX
    # run_example = 2 - download implied volatility data (not in configuration
    # file) from Bloomberg for FX

    run_example = 0

    cache_algo = 'cache_algo_return'

    ###### Download FX volatility quotations from Bloomberg
    if run_example == 1 or run_example == 0:
        ####### Bloomberg examples (you need to have a Bloomberg Terminal
        # installed for this to work!)
        # let's download of 1M ATM data for EURUSD

        # we can use shortcuts given that implied vol surfaces for most major
        # crosses have been defined
        md_request = MarketDataRequest(start_date='01 Jan 2020',
                                       finish_date='15 Feb 2020',
                                       data_source='bloomberg', cut='NYC',
                                       category='fx-implied-vol',
                                       tickers=['EURUSDV1M'],
                                       cache_algo=cache_algo)

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

        # we can also download the whole volatility surface for EURUSD, this
        # way.. without having to define every point!
        md_request = MarketDataRequest(start_date='01 Jan 2020',
                                       finish_date='15 Feb 2020',
                                       data_source='bloomberg',
                                       cut='LDN', category='fx-implied-vol',
                                       tickers=['EURUSD'],
                                       cache_algo=cache_algo)

        df = market.fetch_market(md_request)

        # we can also download the whole all market data for EURUSD for pricing
        # options (vol surface)
        md_request = MarketDataRequest(start_date='01 Jan 2020',
                                       finish_date='15 Feb 2020',
                                       data_source='bloomberg',
                                       cut='LDN', category='fx-vol-market',
                                       tickers=['EURUSD'],
                                       cache_algo=cache_algo)

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

###### Download FX volatility quotations from Bloomberg defining all fields
if run_example == 2 or run_example == 0:
    ####### Bloomberg examples (you need to have a Bloomberg Terminal installed
    # for this to work!)
    # now we define the vendor_tickers and vendor_fields (we don't need to have
    # these in the configuration file)
    # we use NOK/SEK vol quotations, because these haven't been predefined

    # we can use shortcuts given that implied vol surfaces for most major
    # crosses have been defined
    md_request = MarketDataRequest(start_date='01 Jan 2020',
                                   finish_date='15 Feb 2020',
                                   data_source='bloomberg',
                                   tickers=['NOKSEKV1M'],
                                   vendor_tickers=['NOKSEKV1M BGN Curncy'],
                                   fields=['close'], vendor_fields=['PX_LAST'],
                                   cache_algo=cache_algo)

    df = market.fetch_market(md_request)
    print(df.tail(n=10))



from setuptools import setup, find_packages

long_description = """findatapy creates an easy to use Python API to download market data from many sources including 
Quandl, Bloomberg, Yahoo, Google etc. using a unified high level interface. Users can also define their own custom 
tickers, using configuration files. There is also functionality which is particularly useful for those downloading FX market data."""

setup(name="findatapy",
      version="0.1.38",
      description="Market data library",
      author="Saeed Amen",
      author_email="saeed@cuemacro.com",
      license="Apache 2.0",
      long_description=long_description,
      keywords=["pandas", "data", "Bloomberg", "tick", "stocks", "equities"],
      url="https://github.com/cuemacro/findatapy",
      packages=find_packages(),
      include_package_data=True,
      install_requires=["pandas",
                        "twython",
                        "pytz",
                        "requests",
                        "numpy",
                        "pandas_datareader",
                        "alpha_vantage",
                        "eikon",
                        "financepy",
                        "yfinance",
                        "quandl",
                        "chartpy",
                        "statsmodels",
                        "multiprocess",
                        "pathos",
                        "redis",
                        "numba",
                        "pyarrow",
                        "keyring",
                        "openpyxl"],
      extras_require={
          "databento": ["databento"]
      },
      zip_safe=False)



__author__ = "saeedamen"  # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#

from urllib.request import urlopen
from urllib.parse import quote_plus
from urllib.parse import urlencode
from urllib.error import HTTPError

import xml.etree.ElementTree as ET

import pandas as pd

from findatapy.timeseries import Filter, Calculations

from findatapy.market.datavendor import DataVendor
from findatapy.util import LoggerManager

"""Contains implementations of DataVendor for

ALFRED/FRED (free datasource) - DataVendorALFRED

"""

class DataVendorALFRED(DataVendor):
    """Class for reading in data from ALFRED (and FRED) into findatapy library
    (based upon fredapi from https://github.com/mortada/fredapi)

    """

    def __init__(self):
        super(DataVendorALFRED, self).__init__()

    # implement method in abstract superclass
    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)

        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request ALFRED/FRED data")

        data_frame = self.download_daily(md_request_vendor)

        if data_frame is None or data_frame.index is []: return None

        # convert from vendor to findatapy tickers/fields
        if data_frame is not None:
            returned_tickers = data_frame.columns

        if data_frame is not None:
            # tidy up tickers into a format that is more easily translatable
            # we can often get multiple fields returned (even if we don't ask
            # for them!)
            # convert to lower case
            returned_fields = [(x.split('.')[1]) for x in returned_tickers]
            returned_tickers = [(x.split('.')[0]) for x in returned_tickers]

            try:
                fields = self.translate_from_vendor_field(returned_fields,
                                                          md_request)
                tickers = self.translate_from_vendor_ticker(returned_tickers,
                                                            md_request)
            except:
                print('error')

            ticker_combined = []

            for i in range(0, len(fields)):
                ticker_combined.append(tickers[i] + "." + fields[i])

            data_frame.columns = ticker_combined
            data_frame.index.name = 'Date'

        logger.info(
            "Completed request from ALFRED/FRED for " + str(ticker_combined))

        return data_frame

    def download_daily(self, md_request):
        logger = LoggerManager().getLogger(__name__)

        trials = 0

        data_frame_list = []
        data_frame_release = []

        # TODO refactor this code, a bit messy at the moment!
        for i in range(0, len(md_request.tickers)):
            while (trials < 5):
                try:
                    fred = Fred(api_key=md_request.fred_api_key)

                    # acceptable fields: close, actual-release,
                    # release-date-time-full
                    if 'close' in md_request.fields and \
                            'release-date-time-full' in md_request.fields:
                        data_frame = fred.get_series_all_releases(
                            md_request.tickers[i],
                            observation_start=md_request.start_date,
                            observation_end=md_request.finish_date)

                        data_frame = data_frame.rename(
                            columns={
                                "realtime_start":
                                    md_request.tickers[
                                                      i]
                                    + '.release-date-time-full',
                                "date": "Date",
                                "value": md_request.tickers[
                                             i] + '.close'})

                        data_frame = data_frame.sort_values(
                            by=['Date', md_request.tickers[
                                i] + '.release-date-time-full'])
                        data_frame = data_frame.drop_duplicates(
                            subset=['Date'], keep='last')
                        data_frame = data_frame.set_index(['Date'])

                        filter = Filter()
                        data_frame = filter.filter_time_series_by_date(
                            md_request.start_date,
                            md_request.finish_date, data_frame)

                        data_frame_list.append(data_frame)
                    elif 'close' in md_request.fields:

                        data_frame = fred.get_series(
                            series_id=md_request.tickers[i],
                            observation_start=md_request.start_date,
                            observation_end=md_request.finish_date)

                        data_frame = data_frame.to_frame(
                            name=md_request.tickers[i] + '.close')
                        data_frame.index.name = 'Date'

                        data_frame_list.append(data_frame)

                    if 'first-revision' in md_request.fields:
                        data_frame = fred.get_series_first_revision(
                            md_request.tickers[i],
                            observation_start=md_request.start_date,
                            observation_end=md_request.finish_date)

                        data_frame = data_frame.to_frame(
                            name=md_request.tickers[
                                     i] + '.first-revision')
                        data_frame.index.name = 'Date'

                        filter = Filter()
                        data_frame = filter.filter_time_series_by_date(
                            md_request.start_date,
                            md_request.finish_date, data_frame)

                        data_frame_list.append(data_frame)

                    if 'actual-release' in md_request.fields and \
                            'release-date-time-full' in md_request.fields:
                        data_frame = fred.get_series_all_releases(
                            md_request.tickers[i],
                            observation_start=md_request.start_date,
                            observation_end=md_request.finish_date)

                        data_frame = data_frame.rename(
                            columns={
                                "realtime_start": md_request.tickers[
                                                      i]
                                                  + '.release-date-time-full',
                                "date": "Date",
                                "value": md_request.tickers[
                                             i] + '.actual-release'})

                        data_frame = data_frame.sort_values(
                            by=['Date', md_request.tickers[
                                i] + '.release-date-time-full'])
                        data_frame = data_frame.drop_duplicates(
                            subset=['Date'], keep='first')
                        data_frame = data_frame.set_index(['Date'])

                        filter = Filter()
                        data_frame = filter.filter_time_series_by_date(
                            md_request.start_date,
                            md_request.finish_date, data_frame)

                        data_frame_list.append(data_frame)

                    elif 'actual-release' in md_request.fields:
                        data_frame = fred.get_series_first_release(
                            md_request.tickers[i],
                            observation_start=md_request.start_date,
                            observation_end=md_request.finish_date)

                        data_frame = data_frame.to_frame(
                            name=md_request.tickers[
                                     i] + '.actual-release')
                        data_frame.index.name = 'Date'
                        # data_frame = data_frame.rename(columns={"value" :
                        # md_request.tickers[i] + '.actual-release'})

                        filter = Filter()
                        data_frame = filter.filter_time_series_by_date(
                            md_request.start_date,
                            md_request.finish_date, data_frame)

                        data_frame_list.append(data_frame)

                    elif 'release-date-time-full' in md_request.fields:
                        data_frame = fred.get_series_all_releases(
                            md_request.tickers[i],
                            observation_start=md_request.start_date,
                            observation_end=md_request.finish_date)

                        data_frame = data_frame['realtime_start']

                        data_frame = data_frame.to_frame(
                            name=md_request.tickers[
                                     i] + '.release-date-time-full')

                        data_frame.index = data_frame[
                            md_request.tickers[
                                i] + '.release-date-time-full']
                        data_frame = data_frame.sort_index()
                        data_frame = data_frame.drop_duplicates()

                        filter = Filter()
                        data_frame_release.append(
                            filter.filter_time_series_by_date(
                                md_request.start_date,
                                md_request.finish_date,
                                data_frame))

                    break
                except Exception as e:
                    trials = trials + 1
                    logger.info("Attempting... " + str(
                        trials) + " request to download from ALFRED/FRED"
                                + str(e))

            if trials == 5:
                logger.error(
                    "Couldn't download from ALFRED/FRED after several "
                    "attempts!")

        calc = Calculations()

        data_frame1 = calc.join(data_frame_list, how='outer')
        data_frame2 = calc.join(data_frame_release, how='outer')

        data_frame = pd.concat([data_frame1, data_frame2], axis=1)

        return data_frame

class Fred(object):
    """Auxillary class for getting access to ALFRED/FRED directly.

    Based on https://github.com/mortada/fredapi (with minor edits)

    """
    earliest_realtime_start = '1776-07-04'
    latest_realtime_end = '9999-12-31'
    nan_char = '.'
    max_results_per_request = 1000

    def __init__(self,
                 api_key=None,
                 api_key_file=None):
        """Initialize the Fred class that provides useful functions to query the Fred dataset. You need to specify a valid
        API key in one of 3 ways: pass the string via api_key, or set api_key_file to a file with the api key in the
        first line, or set the environment variable 'FRED_API_KEY' to the value of your api key. You can sign up for a
        free api key on the Fred website at http://research.stlouisfed.org/fred2/
        """
        self.api_key = None
        if api_key is not None:
            self.api_key = api_key
        elif api_key_file is not None:
            f = open(api_key_file, 'r')
            self.api_key = f.readline().strip()
            f.close()
        else:
            self.api_key = os.environ.get('FRED_API_KEY')
        self.root_url = 'https://api.stlouisfed.org/fred'

        if self.api_key is None:
            import textwrap

            raise ValueError(textwrap.dedent("""\
                    You need to set a valid API key. You can set it in 3 ways:
                    pass the string with api_key, or set api_key_file to a
                    file with the api key in the first line, or set the
                    environment variable 'FRED_API_KEY' to the value of your
                    api key. You can sign up for a free api key on the Fred
                    website at http://research.stlouisfed.org/fred2/"""))

    def __fetch_data(self, url):
        """Helper function for fetching data given a request URL
        """
        try:
            response = urlopen(url)
            root = ET.fromstring(response.read())
        except HTTPError as exc:
            root = ET.fromstring(exc.read())
            raise ValueError(root.get('message'))
        return root

    def _parse(self, date_str, format='%Y-%m-%d'):
        """Helper function for parsing FRED date string into datetime

        """
        from pandas import to_datetime
        rv = to_datetime(date_str, format=format)
        if hasattr(rv, 'to_datetime'):
            rv = rv.to_pydatetime()  # to_datetime is depreciated
        return rv

    def get_series_info(self, series_id):
        """Get information about a series such as its title, frequency, observation start/end dates, units, notes, etc.

        Parameters
        ----------
        series_id : str
            Fred series id such as 'CPIAUCSL'

        Returns
        -------
        info : Series
            a pandas Series containing information about the Fred series
        """
        url = "%s/series?series_id=%s&api_key=%s" % (self.root_url, series_id,
                                                     self.api_key)
        root = self.__fetch_data(url)
        if root is None:
            raise ValueError('No info exists for series id: ' + series_id)
        from pandas import Series
        info = Series(root.getchildren()[0].attrib)
        return info

    def get_series(self, series_id, observation_start=None,
                   observation_end=None, **kwargs):
        """Get data for a Fred series id. This fetches the latest known data, and is equivalent to get_series_latest_release()

        Parameters
        ----------
        series_id : str
            Fred series id such as 'CPIAUCSL'
        observation_start : datetime or datetime-like str such as '7/1/2014', optional
            earliest observation date
        observation_end : datetime or datetime-like str such as '7/1/2014', optional
            latest observation date
        kwargs : additional parameters
            Any additional parameters supported by FRED. You can see https://api.stlouisfed.org/docs/fred/series_observations.html for the full list

        Returns
        -------
        data : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        url = "%s/series/observations?series_id=%s&api_key=%s" % (
        self.root_url,
        series_id,
        self.api_key)
        from pandas import to_datetime, Series

        if observation_start is not None:
            observation_start = to_datetime(observation_start, errors='raise')
            url += '&observation_start=' + observation_start.strftime(
                '%Y-%m-%d')
        if observation_end is not None:
            observation_end = to_datetime(observation_end, errors='raise')
            url += '&observation_end=' + observation_end.strftime('%Y-%m-%d')

        if kwargs is not None:
            url += '&' + urlencode(kwargs)

        root = self.__fetch_data(url)
        if root is None:
            raise ValueError('No data exists for series id: ' + series_id)
        data = {}
        for child in root.iter():
            val = child.get('value')
            if val is None:
                float('NaN')
            elif val == self.nan_char:
                val = float('NaN')
            else:
                val = float(val)
            data[self._parse(child.get('date'))] = val
        return Series(data)

    def get_series_latest_release(self, series_id, observation_start=None,
                                  observation_end=None):
        """Get data for a Fred series id. This fetches the latest known data, and is equivalent to get_series()

        Parameters
        ----------
        series_id : str
            Fred series id such as 'CPIAUCSL'

        Returns
        -------
        info : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        return self.get_series(series_id, observation_start=observation_start,
                               observation_end=observation_end)

    def get_series_first_release(self, series_id, observation_start=None,
                                 observation_end=None):
        """Get first-release data for a Fred series id.

        This ignores any revision to the data series. For instance,
        The US GDP for Q1 2014 was first released to be 17149.6, and then later revised to 17101.3, and 17016.0.
        This will ignore revisions after the first release.

        Parameters
        ----------
        series_id : str
            Fred series id such as 'GDP'

        Returns
        -------
        data : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        df = self.get_series_all_releases(series_id,
                                          observation_start=observation_start,
                                          observation_end=observation_end)
        first_release = df.groupby('date').head(1)
        data = first_release.set_index('date')['value']
        return data

    def get_series_first_revision(self, series_id, observation_start=None,
                                  observation_end=None):
        """Get first-revision data for a Fred series id.

        This will give the first revision to the data series. For instance,
        The US GDP for Q1 2014 was first released to be 17149.6, and then later revised to 17101.3, and 17016.0.
        This will take the first revision ie. 17101.3 in this case.

        Parameters
        ----------
        series_id : str
            Fred series id such as 'GDP'

        Returns
        -------
        data : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        df = self.get_series_all_releases(series_id,
                                          observation_start=observation_start,
                                          observation_end=observation_end)
        first_revision = df.groupby('date').head(2)
        data = first_revision.set_index('date')['value']
        data = data[~data.index.duplicated(keep='last')]

        return data

    def get_series_as_of_date(self, series_id, as_of_date):
        """Get latest data for a Fred series id as known on a particular date.

        This includes any revision to the data series
        before or on as_of_date, but ignores any revision on dates after as_of_date.

        Parameters
        ----------
        series_id : str
            Fred series id such as 'GDP'
        as_of_date : datetime, or datetime-like str such as '10/25/2014'
            Include data revisions on or before this date, and ignore revisions afterwards

        Returns
        -------
        data : Series
            a Series where each index is the observation date and the value is the data for the Fred series
        """
        from pandas import to_datetime
        as_of_date = to_datetime(as_of_date)
        df = self.get_series_all_releases(series_id)
        data = df[df['realtime_start'] <= as_of_date]
        return data

    def get_series_all_releases(self, series_id, observation_start=None,
                                observation_end=None):
        """Get all data for a Fred series id including first releases and all revisions.

        This returns a DataFrame
        with three columns: 'date', 'realtime_start', and 'value'. For instance, the US GDP for Q4 2013 was first released
        to be 17102.5 on 2014-01-30, and then revised to 17080.7 on 2014-02-28, and then revised to 17089.6 on
        2014-03-27. You will therefore get three rows with the same 'date' (observation date) of 2013-10-01 but three
        different 'realtime_start' of 2014-01-30, 2014-02-28, and 2014-03-27 with corresponding 'value' of 17102.5, 17080.7
        and 17089.6

        Parameters
        ----------
        series_id : str
            Fred series id such as 'GDP'

        Returns
        -------
        data : DataFrame
            a DataFrame with columns 'date', 'realtime_start' and 'value' where 'date' is the observation period and 'realtime_start'
            is when the corresponding value (either first release or revision) is reported.
        """
        url = "%s/series/observations?series_id=%s&api_key=%s&realtime_start=%s&realtime_end=%s" % (
        self.root_url,
        series_id,
        self.api_key,
        self.earliest_realtime_start,
        self.latest_realtime_end)

        from pandas import to_datetime

        if observation_start is not None:
            observation_start = to_datetime(observation_start, errors='raise')
            url += '&observation_start=' + observation_start.strftime(
                '%Y-%m-%d')
        if observation_end is not None:
            observation_end = to_datetime(observation_end, errors='raise')
            url += '&observation_end=' + observation_end.strftime('%Y-%m-%d')

        root = self.__fetch_data(url)
        if root is None:
            raise ValueError('No data exists for series id: ' + series_id)
        data = {}
        i = 0
        for child in root.iter():
            val = child.get('value')
            if val is None:
                val = float("NaN")
            elif val == self.nan_char:
                val = float('NaN')
            else:
                val = float(val)
            realtime_start = self._parse(child.get('realtime_start'))
            # realtime_end = self._parse(child.get('realtime_end'))
            date = self._parse(child.get('date'))

            data[i] = {'realtime_start': realtime_start,
                       # 'realtime_end': realtime_end,
                       'date': date,
                       'value': val}
            i += 1
        from pandas import DataFrame
        data = DataFrame(data).T
        return data

    def get_series_vintage_dates(self, series_id):
        """Get a list of vintage dates for a series.

        Vintage dates are the dates in history when a series' data values were revised or new data values were released.

        Parameters
        ----------
        series_id : str
            Fred series id such as 'CPIAUCSL'

        Returns
        -------
        dates : list
            list of vintage dates
        """
        url = "%s/series/vintagedates?series_id=%s&api_key=%s" % (
        self.root_url,
        series_id,
        self.api_key)
        root = self.__fetch_data(url)
        if root is None:
            raise ValueError(
                'No vintage date exists for series id: ' + series_id)
        dates = []
        for child in root.getchildren():
            dates.append(self._parse(child.text))
        return dates

    def __do_series_search(self, url):
        """Helper function for making one HTTP request for data, and parsing the returned results into a DataFrame
        """
        root = self.__fetch_data(url)

        series_ids = []
        data = {}

        num_results_returned = 0  # number of results returned in this HTTP request
        num_results_total = int(
            root.get(
                'count'))  # total number of results, this can be larger than number of results returned
        for child in root.getchildren():
            num_results_returned += 1
            series_id = child.get('id')
            series_ids.append(series_id)
            data[series_id] = {"id": series_id}
            fields = ["realtime_start", "realtime_end", "title",
                      "observation_start", "observation_end",
                      "frequency", "frequency_short", "units", "units_short",
                      "seasonal_adjustment",
                      "seasonal_adjustment_short", "last_updated",
                      "popularity", "notes"]
            for field in fields:
                data[series_id][field] = child.get(field)

        if num_results_returned > 0:
            from pandas import DataFrame
            data = DataFrame(data, columns=series_ids).T
            # parse datetime columns
            for field in ["realtime_start", "realtime_end",
                          "observation_start", "observation_end",
                          "last_updated"]:
                data[field] = data[field].apply(self._parse, format=None)
            # set index name
            data.index.name = 'series id'
        else:
            data = None
        return data, num_results_total

    def __get_search_results(self, url, limit, order_by, sort_order):
        """Helper function for getting search results up to specified limit on the number of results. The Fred HTTP API
        truncates to 1000 results per request, so this may issue multiple HTTP requests to obtain more available data.
        """

        order_by_options = ['search_rank', 'series_id', 'title', 'units',
                            'frequency',
                            'seasonal_adjustment', 'realtime_start',
                            'realtime_end', 'last_updated',
                            'observation_start', 'observation_end',
                            'popularity']
        if order_by is not None:
            if order_by in order_by_options:
                url = url + '&order_by=' + order_by
            else:
                raise ValueError(
                    '%s is not in the valid list of order_by options: %s' % (
                    order_by, str(order_by_options)))

        sort_order_options = ['asc', 'desc']
        if sort_order is not None:
            if sort_order in sort_order_options:
                url = url + '&sort_order=' + sort_order
            else:
                raise ValueError(
                    '%s is not in the valid list of sort_order options: %s' % (
                    sort_order, str(sort_order_options)))

        data, num_results_total = self.__do_series_search(url)
        if data is None:
            return data

        if limit == 0:
            max_results_needed = num_results_total
        else:
            max_results_needed = limit

        if max_results_needed > self.max_results_per_request:
            for i in range(1,
                           max_results_needed // self.max_results_per_request + 1):
                offset = i * self.max_results_per_request
                next_data, _ = self.__do_series_search(
                    url + '&offset=' + str(offset))
                data = data.append(next_data)
        return data.head(max_results_needed)

    def search(self, text, limit=1000, order_by=None, sort_order=None):
        """Do a fulltext search for series in the Fred dataset.

        Returns information about matching series in a DataFrame.

        Parameters
        ----------
        text : str
            text to do fulltext search on, e.g., 'Real GDP'
        limit : int, optional
            limit the number of results to this value. If limit is 0, it means fetching all results without limit.
        order_by : str, optional
            order the results by a criterion. Valid options are 'search_rank', 'series_id', 'title', 'units', 'frequency',
            'seasonal_adjustment', 'realtime_start', 'realtime_end', 'last_updated', 'observation_start', 'observation_end',
            'popularity'
        sort_order : str, optional
            sort the results by ascending or descending order. Valid options are 'asc' or 'desc'

        Returns
        -------
        info : DataFrame
            a DataFrame containing information about the matching Fred series
        """
        url = "%s/series/search?search_text=%s&api_key=%s" % (self.root_url,
                                                              quote_plus(text),
                                                              self.api_key)
        info = self.__get_search_results(url, limit, order_by, sort_order)
        return info

    def search_by_release(self, release_id, limit=0, order_by=None,
                          sort_order=None):
        """Search for series that belongs to a release id.

        Returns information about matching series in a DataFrame.

        Parameters
        ----------
        release_id : int
            release id, e.g., 151
        limit : int, optional
            limit the number of results to this value. If limit is 0, it means fetching all results without limit.
        order_by : str, optional
            order the results by a criterion. Valid options are 'search_rank', 'series_id', 'title', 'units', 'frequency',
            'seasonal_adjustment', 'realtime_start', 'realtime_end', 'last_updated', 'observation_start', 'observation_end',
            'popularity'
        sort_order : str, optional
            sort the results by ascending or descending order. Valid options are 'asc' or 'desc'

        Returns
        -------
        info : DataFrame
            a DataFrame containing information about the matching Fred series
        """
        url = "%s/release/series?release_id=%d&&api_key=%s" % (self.root_url,
                                                               release_id,
                                                               self.api_key)
        info = self.__get_search_results(url, limit, order_by, sort_order)
        if info is None:
            raise ValueError(
                'No series exists for release id: ' + str(release_id))
        return info

    def search_by_category(self, category_id, limit=0, order_by=None,
                           sort_order=None):
        """Search for series that belongs to a category id.

        Returns information about matching series in a DataFrame.

        Parameters
        ----------
        category_id : int
            category id, e.g., 32145
        limit : int, optional
            limit the number of results to this value. If limit is 0, it means fetching all results without limit.
        order_by : str, optional
            order the results by a criterion. Valid options are 'search_rank', 'series_id', 'title', 'units', 'frequency',
            'seasonal_adjustment', 'realtime_start', 'realtime_end', 'last_updated', 'observation_start', 'observation_end',
            'popularity'
        sort_order : str, optional
            sort the results by ascending or descending order. Valid options are 'asc' or 'desc'

        Returns
        -------
        info : DataFrame
            a DataFrame containing information about the matching Fred series
        """
        url = "%s/category/series?category_id=%d&api_key=%s" % (self.root_url,
                                                                category_id,
                                                                self.api_key)
        info = self.__get_search_results(url, limit, order_by, sort_order)
        if info is None:
            raise ValueError(
                'No series exists for category id: ' + str(category_id))
        return info



__author__ = "saeedamen"  # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#


class CommonMan(object):

    def prefix_list(self, list, prefix):
        out = []

        for li in list:
            out.append(prefix + li)

        return out

    def postfix_list(self, list, postfix):
        out = []

        for li in list:
            out.append(li + postfix)

        return out


__author__ = "saeedamen"  # Saeed Amen

#
# Copyright 2024 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#

from findatapy.market.datavendor import DataVendor

# Databento is an optional dependency
try:
    import databento as db
except:
    pass

class DataVendorDatabento(DataVendor):

    def __init__(self):
        super(DataVendorDatabento, self).__init__()

    def load_ticker(self, md_request):
        logger = LoggerManager().getLogger(__name__)
        md_request_vendor = self.construct_vendor_md_request(
            md_request)

        logger.info("Request Databento data")

        data_frame = self.download_daily(md_request_vendor)

        if data_frame is None or data_frame.index is []:
            return None

        # Convert from vendor to findatapy tickers/fields
        if data_frame is not None:
            if len(md_request.fields) == 1:
                data_frame.columns = data_frame.columns.str.cat(
                    md_request.fields * len(data_frame.columns),
                    sep='.')
            else:
                logger.warning("Inconsistent number of fields and tickers.")
                data_frame.columns = data_frame.columns.str.cat(
                    md_request.fields, sep='.')
            data_frame.index.name = 'Date'

        logger.info("Completed request from Databento")

        return data_frame

    def download_daily(self, md_request):
        logger = LoggerManager().getLogger(__name__)
        trials = 0

        data_frame = None

        if md_request.freq == "daily":
            schema ="ohlcv-1d"
        elif md_request.freq == "intraday":
            schema = "ohlcv-1m"

        while trials < 5:
            try:
                client = db.Historical(md_request.databento_api_key)

                data = client.timeseries.get_range(
                    dataset=md_request.data_source.split("-")[1],
                    symbols=md_request.vendor_tickers,
                    schema=schema,
                    start=pd.Timestamp(md_request.start_date),
                    end=pd.Timestamp(md_request.finish_date),
                    # limit=1,
                )
                data_frame = data.to_df()
            except:
                trials = trials + 1
                logger.info(f"Attempting... {str(trials)} request to download from Databento")

        data_frame = data_frame.set_index('symbol', append=True).unstack(level='symbol')
        fields = df.columns.levels[0]
        tickers = df.columns.levels[1]

        new_cols = []

        for fi in fields:
            for ti in tickers:
                new_cols.append(ti + "." + fi)

        data_frame.columns = new_cols
        data_frame.index.name = "Date"

        if trials == 5:
            logger.error("Couldn't download from Databento after several attempts!")

        return data_frame



__author__ = "saeedamen"  # Saeed Amen

#
# Copyright 2024 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on a "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#

import datetime
import time

if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool;

    SwimPool()

    # This is the new ArcticDB, which uses disk based storage, as opposed
    # to a database engine like MongoDB as a backend
    from findatapy.market import Market, MarketDataRequest, \
        MarketDataGenerator, IOEngine

    market = Market(market_data_generator=MarketDataGenerator())

    from findatapy.market import Market, MarketDataRequest

    # In this case we are saving predefined tick tickers to disk, and then reading back
    from findatapy.market.ioengine import IOEngine

    md_request_download = MarketDataRequest(
        start_date="04 Jan 2021",
        finish_date="06 Jan 2021",
        category="fx",
        data_source="dukascopy",
        freq="tick",
        tickers=["USDJPY"],
        fields=["bid", "ask", "bidv", "askv"],
        data_engine=None
    )

    market = Market()

    df_tick_earlier = market.fetch_market(md_request=md_request_download)

    io = IOEngine()

    local_storage = True

    # To switch between local storage or s3, it's a matter of changing the
    # connection string (also you need to make sure your AWS S3 authentication is set etc.)
    if local_storage:
        arcticdb_conn_str = "arcticdb:lmdb://tempdatabase?map_size=2GB"
    else:
        # https://docs.arcticdb.io/latest/#s3-configuration gives more details
        # Note we need to prefix arcticdb: to the front so findatapy
        # knows what backend engine to use
        region = "eu-west-2"
        bucket_name = "burger_king_whopper" # Not sure, if this name is taken :-)
        path_prefix = "test"

        arcticdb_conn_str = f"arcticdb:s3s://s3.{region}.amazonaws.com:{bucket_name}?path_prefix={path_prefix}&aws_auth=true"

    # Set various parameters to govern how we write to ArcticDB, to use
    # versioning
    arcticdb_dict = {
        # If this is set to true removes previous versions (so we only record
        # the final version). Not pruning versions will take more disk space
        "prune_previous_versions": False,

        # Do we want to append to existing records or write
        # If you attempt to append with an overlapping chunk, you'll
        # get an assertion failure, "update" allows you to change existing data
        "write_style": "write", # "write" / "append" / "update"

        # If set to true will remove any existing library, before writing
        "force_create_library": False,

        # This enables us to take advantage of ArcticDB's filtering of columns/dates
        # otherwise we would download the full dataset, and then filter
        # in Pandas
        "allow_on_disk_filter": True,

        # You can also specify your own custom queries for ArcticDB
        "query_builder": None
    }

    md_request_download.arcticdb_dict = arcticdb_dict

    IOEngine().write_time_series_cache_to_disk(data_frame=df_tick_earlier,
                                               engine=arcticdb_conn_str,
                                               md_request=md_request_download)
    earlier_download_time = datetime.datetime.now().utcnow()

    # Let's read directly from ArcticDB without the Market wrapper using IOEngine
    # This would also be the symbol you can use if you want to directly
    # interact with Arctic
    symbol = "backtest.fx.dukascopy.tick.NYC.USDJPY"
    df_read_tick = IOEngine().read_time_series_cache_from_disk(
        symbol, engine=arcticdb_conn_str)

    print(df_read_tick)

    # Now download a second set of data and write it for "append"
    # Note: we'll get an assertion error, if we try to append before the end
    # of the existing time series on disk
    md_request_download.start_date = "06 Jan 2021"
    md_request_download.finish_date = "07 Jan 2021"
    md_request_download.arcticdb_dict["write_style"] = "append"

    df_tick_later = market.fetch_market(md_request=md_request_download)
    IOEngine().write_time_series_cache_to_disk(data_frame=df_tick_later,
                                               engine=arcticdb_conn_str,
                                               md_request=md_request_download)

    # This time we use the Market wrapper to download data
    # Given we don't specify an "as_of" property, we'll get the later version
    later_download_time = datetime.datetime.now().utcnow()

    md_request_local_cache = MarketDataRequest(
        md_request=md_request_download
    )

    md_request_local_cache.start_date = "04 Jan 2021 10:00"
    md_request_local_cache.finish_date = "06 Jan 2021 14:00"
    md_request_local_cache.data_engine = arcticdb_conn_str
    md_request_local_cache.cache_algo = "cache_algo_return"

    df_read_tick = Market().fetch_market(md_request=md_request_local_cache)

    # We should see the 1st write and 2nd append combined, ie. latest write
    print("No as_of specified, so we'll get the latest write!")
    print(df_read_tick)

    # Let's instead take the first vintage
    md_request_local_cache.as_of = earlier_download_time
    df_read_tick = Market().fetch_market(md_request=md_request_local_cache)

    # We should only see the earlier vintage
    print("See the earlier vintage write!")
    print(df_read_tick)

    # We can also specify the later write time
    md_request_local_cache.as_of = later_download_time
    df_read_tick = Market().fetch_market(md_request=md_request_local_cache)

    # We should only see the latest vintage
    print("See the latest vintage write!")
    print(df_read_tick)

    # Finally let's try doing an update of an existing continuous chunk
    # Note: if part of our update ends up being before/after the existing
    # dataset, it will fail
    md_request_download.start_date = "06 Jan 2021"
    md_request_download.finish_date = "07 Jan 2021"
    md_request_download.arcticdb_dict["write_style"] = "update"

    df_tick_later = market.fetch_market(md_request=md_request_download)

    # Modify the date, so we can see the obvious difference!
    df_tick_later = df_tick_later * 10.0
    IOEngine().write_time_series_cache_to_disk(data_frame=df_tick_later,
                                               engine=arcticdb_conn_str,
                                               md_request=md_request_download)

    md_request_local_cache.start_date = "06 Jan 2021 10:00"
    md_request_local_cache.finish_date = "08 Jan 2021 14:00"
    md_request_local_cache.as_of = None
    df_read_updated_tick = Market().fetch_market(md_request=md_request_local_cache)

    print("Updated tick (should be 10 larger!)")
    print(df_read_updated_tick)


